@@ -8,8 +8,9 @@ from .types import Interval, TimeInterval
 
 
 class IntervalIndex(BaseIndex):
+    
     __slots__ = BaseIndex.__slots__ + ("closed",)
-
+    
     def __set_name__(self, owner, name):
         super().__set_name__(owner, name)
 
@@ -54,5 +55,7 @@ class IntervalIndex(BaseIndex):
 
     def from_pandas(self, label):
         if isinstance(label, pd.Interval):
-            label = Interval(left=label.left, right=label.right)
+            type_ = type(label.left)
+            interval_class = Interval[type_]
+            label = interval_class(left=label.left, right=label.right)
         return label
