@@ -1,13 +1,15 @@
 import datetime
-from typing import Mapping, TypeVar
+from typing import ClassVar, Mapping, TypeVar
 
 import pydantic
-from pydantic import BaseModel
+from pydantic import BaseModel, root_validator
 
 LabelType = TypeVar("LabelType", int, str, datetime.datetime)
 
 
 class Interval(BaseModel):
+    _resolution: ClassVar = None
+
     left: LabelType
     right: LabelType = None
 
@@ -17,6 +19,8 @@ class Interval(BaseModel):
 
     @classmethod
     def validate_field(cls, v, field):
+        if isinstance(v, cls):
+            return v
 
         if isinstance(v, tuple):
             left, right = v
@@ -28,10 +32,11 @@ class Interval(BaseModel):
             left = v.left
             right = v.right
         else:
-            left, right = v, v
+            if cls._resolution is not None:
+                left, right = v, v
+            else:
+                left, right = v, v
 
-        if right is not None and left > right:
-            left, right = right, left
         return cls(left=left, right=right)
 
     def __class_getitem__(cls, type_):
@@ -40,13 +45,31 @@ class Interval(BaseModel):
         if issubclass(type_, datetime.datetime):
             return TimeInterval
         raise TypeError(type_)
+        
+    @root_validator
+    def check_non_zero_length(cls, values):
+        left, right = values.get('left'), values.get('right')
+
+        if right is not None and left > right:
+            left, right = right, left
+
+        if (right - left ) < cls._resolution:
+            left = left - cls._resolution
+
+        values['left'] = left
+        values['right'] = right 
+        return values
 
 
 class IntegerInterval(Interval):
+    _resolution = 1
+
     left = pydantic.conint(ge=0, lt=int(2**32 - 1))
     right = pydantic.conint(ge=0, lt=int(2**32 - 1))
 
 
 class TimeInterval(Interval):
+    _resolution = datetime.timedelta(microseconds=1000)
+
     left: datetime.datetime
     right: datetime.datetime = None
