@@ -1,8 +1,10 @@
+from loguru import logger
+
 from ..utils import singledispatch
 from .base import DatasourceInterface
 from .mongo import MongoInterface
 from .pandas import PandasInterface
-from .http import HttpInterface
+from .rest import RestInterface
 
 
 def get_interface(source, **kwargs) -> DatasourceInterface:
