@@ -1,5 +1,7 @@
 from itertools import product
+from typing import List, Union
 from warnings import warn
+from loguru import logger
 
 import pandas as pd
 from pydantic import BaseModel
@@ -25,10 +27,16 @@ try:
             self.collection = collection
             self.aggregations = aggregations
 
-        def execute(self):
+        def execute(self, limit=None, skip=None):
+            logger.debug('Executing multi mongo aggregation.')
             results = []
             for agg in self.aggregations:
-                results.extend(agg.apply(self.collection))
+                results.extend(agg.execute(self.collection, limit=limit, skip=skip))
+            skip = 0 if skip is None else skip
+
+            if limit is not None:
+                return results[skip:limit+skip]
+
             return results
 
         def logical_or(self, other: "MultiMongoAggregation"):
@@ -48,20 +56,102 @@ try:
     class MongoAggregation(BaseDataQuery):
         pipeline: list
 
-        def __init__(self, collection: Collection, pipeline: list):
+        def __init__(self, index, labels, collection: Collection, pipeline: list):
             if not isinstance(collection, Collection):
                 raise TypeError(
                     f"collection must be a pymongo Collection, got {type(collection)}."
                 )
-
+            self.index = index
+            self.labels = labels
             self.collection = collection
             self.pipeline = pipeline
 
-        def execute(self):
-            return list(self.collection.aggregate(self.pipeline, allowDiskUse=True))
+        def execute(self, limit: int = None, skip: int = None):
+            pipeline = list(self.pipeline)
+
+            if isinstance(skip, int):
+                raw_skip = skip * self.index.DOCS_PER_LABEL
+                pipeline.append({"$skip": raw_skip})
+
+            if isinstance(limit, int):
+                raw_limit = limit * self.index.DOCS_PER_LABEL
+                raw_limit = int(raw_limit)
+                pipeline.append({"$limit": raw_limit})
+            
+            pipeline.append({"$project": { "_id": 0}})
+
+            logger.debug(f'Executing mongo aggregation: {pipeline}.')
+
+            docs = list(self.collection.aggregate(pipeline, allowDiskUse=True))
+
+            docs = self.index.reduce(docs, self.labels)
+            return docs
+
+        def unique(self, fields: Union[str, List[str]]):
+            if isinstance(fields, str):
+                fields = [fields]
+            results = {}
+            for field in fields:
+                pipeline = list(self.pipeline)
+                pipeline.append({
+                    "$group": {
+                        "_id": "$" + field,
+                        'first': { '$first':  "$" + field },
+                    }
+                    })
+                    
+                results[field] = [doc['first'] for doc in self.collection.aggregate(pipeline, allowDiskUse=True)]
+            if len(fields) == 1:
+                return results[fields[0]]
+            return results
+        
+        def max(self, fields: Union[str, List[str]]):
+            if isinstance(fields, str):
+                fields = [fields]
+            results = {}
+            for field in fields:
+                pipeline = list(self.pipeline)
+                pipeline.append({
+                    "$sort": { field: -1}
+                })
+                pipeline.append({"$limit": 1})
+                pipeline.append({"$project": { "_id": 0}})
+                results[field] = next(self.collection.aggregate(pipeline, allowDiskUse=True))[field]
+
+            if len(fields) == 1:
+                return results[fields[0]]
+            return results
+
+        def min(self, fields: Union[str, List[str]]):
+            if isinstance(fields, str):
+                fields = [fields]
+            results = {}
+            for field in fields:
+                pipeline = list(self.pipeline)
+                pipeline.append({
+                    "$sort": { field: 1}
+                })
+                pipeline.append({"$limit": 1})
+                pipeline.append({"$project": { "_id": 0}})
+                results[field] = next(self.collection.aggregate(pipeline, allowDiskUse=True))[field]
+
+            if len(fields) == 1:
+                return results[fields[0]]
+            return results
+
+        def count(self):
+            pipeline = list(self.pipeline)
+            pipeline.append({"$count": "count"})
+            result = next(self.collection.aggregate(pipeline, allowDiskUse=True))
+            return result.get('count', 0)
 
         def logical_and(self, other):
-            return MongoAggregation(self.pipeline + other.pipeline)
+            index = MultiIndex(self.index, other.index)
+            labels = dict(self.labels, **other.labels)
+            return MongoAggregation(index,
+                                    labels,
+                                    self.collection, 
+                                    self.pipeline + other.pipeline)
 
         def logical_or(self, other):
             if isinstance(other, MongoAggregation):
@@ -103,39 +193,38 @@ try:
                 f"{self.__class__.__name__} does not support {type(index)} indexes."
             )
 
-        def simple_multi_query(self, indexes, labels):
+        def simple_multi_query(self, index, labels):
             pipeline = []
-            names = [idx.name for idx in indexes]
-            for idx, label in zip(indexes, labels):
-                others = [name for name in names if name != idx.name]
+
+            for idx, label in zip(index.indexes, labels.values()):
+                others = [name for name in index.names if name != idx.name]
                 agg = self.compile_query(idx, label, others=others)
                 pipeline.extend(agg.pipeline)
-            return MongoAggregation(self.source, pipeline)
+            return MongoAggregation(index, labels, self.source, pipeline)
 
-        def product_multi_query(self, indexes, labels):
+        def product_multi_query(self, index, labels):
             labels = [label if isinstance(label, list) else [label] for label in labels]
             aggs = []
             for label_vals in product(*labels):
-                agg = self.simple_multi_query(indexes, label_vals)
+                agg = self.simple_multi_query(index.indexes, label_vals)
                 aggs.append(agg)
             return MultiMongoAggregation(self.source, aggs)
 
         @compile_query.register(list)
         @compile_query.register(tuple)
         @compile_query.register(MultiIndex)
-        def multi_query(self, indexes, labels):
-            if isinstance(indexes, MultiIndex):
-                indexes = indexes.indexes
-                labels = labels.values()
-
-            # if any([isinstance(l, list) for l in labels]):
-            #     return self.product_multi_query(indexes, labels)
-
-            return self.simple_multi_query(indexes, labels)
+        def multi_query(self, index, labels):
+            logger.debug('Building mongo multi-query for index: '
+                        f'{index} with labels: {labels}')
+            if not isinstance(index, MultiIndex):
+                index = MultiIndex(*index)
+            return self.simple_multi_query(index, labels)
 
         @compile_query.register(Index)
         @compile_query.register(str)
         def get_simple_query(self, index, label, others=None):
+            logger.debug('Building mongo simple-query for index: '
+                        f'{index} with label: {label}')
             name = index.name if isinstance(index, Index) else index
 
             if isinstance(label, slice):
@@ -170,23 +259,30 @@ try:
             if label is not None:
                 pipeline.append({"$match": match})
 
-            return MongoAggregation(self.source, pipeline)
+            labels = {name: label}
+
+            return MongoAggregation(index, labels, self.source, pipeline)
 
         @compile_query.register(InterpolatingIndex)
         def build_interpolation_query(self, index, label, others=None):
             """For interpolation we match the values directly before and after
             the value of interest. For each value we take the closest document on either side.
             """
+            logger.debug('Building mongo interpolating-query for index: '
+                        f'{index} with label: {label}')
             if label is None:
-                return MongoAggregation(self.source, [])
+                labels = {index.name: label}
+                return MongoAggregation(index, labels, self.source, [])
+            limit = 1 if others is None else others
 
             if not isinstance(label, list):
                 pipelines = dict(
-                    before=mongo_before_query(index.name, label, limit=others),
-                    after=mongo_after_query(index.name, label, limit=others),
+                    before=mongo_before_query(index.name, label, limit=limit),
+                    after=mongo_after_query(index.name, label, limit=limit),
                 )
                 pipeline = merge_pipelines(pipelines)
-                return MongoAggregation(self.source, pipeline)
+                labels = {index.name: label}
+                return MongoAggregation(index, labels, self.source, pipeline)
 
             pipelines = {
                 f"agg{i}": mongo_closest_query(index.name, value)
@@ -194,16 +290,24 @@ try:
             }
             pipeline = merge_pipelines(pipelines)
 
-            return MongoAggregation(self.source, pipeline)
+            labels = {index.name: label}
+
+            return MongoAggregation(index, labels, self.source, pipeline)
 
         @compile_query.register(IntervalIndex)
-        def build_interval_query(self, index, intervals, others=None):
+        def build_interval_query(self, index, label, others=None):
             """Query overlaping documents with given interval, supports multiple
             intervals as well as zero length intervals (left==right)
             multiple overlap queries are joined with the $or operator
             """
-            if not isinstance(intervals, list):
-                intervals = [intervals]
+
+            logger.debug('Building mongo interval-query for index: '
+                        f'{index} with label: {label}')
+
+            if isinstance(label, list):
+                intervals = label
+            else:
+                intervals = [label]
 
             queries = []
             for interval in intervals:
@@ -225,22 +329,14 @@ try:
                             "$or": queries,
                         }
                     },
-                    {
-                        "$project": {
-                            "_id": 0,
-                        },
-                    },
+                    
                 ]
             else:
-                pipeline = [
-                    {
-                        "$project": {
-                            "_id": 0,
-                        },
-                    }
-                ]
+                pipeline = []
 
-            return MongoAggregation(self.source, pipeline)
+            labels = {index.name: intervals}
+
+            return MongoAggregation(index, labels, self.source, pipeline)
 
         def insert(self, doc):
             """We want the client logic to be agnostic to
@@ -265,9 +361,8 @@ try:
                 return doc
             except Exception as e:
                 raise InsertionError(f"Mongodb has rejected this insertion:\n {e} ")
-
-        def insert_many(self, docs):
-            raise NotImplementedError
+        
+        update = insert
 
         def ensure_index(self, names, order=pymongo.ASCENDING):
             self.source.ensure_index([(name, order) for name in names])
@@ -297,8 +392,8 @@ def mongo_overlap_query(index, interval):
     # Set the appropriate operators depending on if the interval
     # is closed on one side or both
     closed = getattr(index, "closed", "right")
-    gt_op = "$gte" if closed in ["right", "both"] else "$gt"
-    lt_op = "$lte" if closed in ["left", "both"] else "$lt"
+    gt_op = "$gte" if closed == "both" else "$gt"
+    lt_op = "$lte" if closed == "both" else "$lt"
 
     # handle different kinds of interval definitions
     if isinstance(interval, tuple):
@@ -373,6 +468,7 @@ def mongo_grouped_before_query(name, value, groups):
             # make the documents the new root, discarding the groupby value
             "$replaceRoot": {"newRoot": "$doc"},
         },
+        
     ]
 
 
@@ -417,7 +513,7 @@ def mongo_closest_query(name, value):
         },
         {
             # drop the extra fields, they are no longer needed
-            "$project": {"_diff": 0, "_after": 0, "_id": 0},
+            "$project": {"_diff": 0, "_after": 0},
         },
     ]
 
@@ -442,10 +538,6 @@ def merge_pipelines(pipelines):
         # move list of documents to the root of the result
         # so we just get a nice list of documents
         {"$replaceRoot": {"newRoot": "$union"}},
-        {
-            "$project": {
-                "_id": 0,
-            },
-        },
+        
     ]
     return pipeline
