@@ -1,5 +1,6 @@
+from loguru import logger
 from datetime import datetime
-from typing import Any, List
+from typing import Any, List, Union
 
 import pandas as pd
 
@@ -9,20 +10,79 @@ from .base import BaseDataQuery, DatasourceInterface
 
 
 class PandasBaseQuery(BaseDataQuery):
-    def __init__(self, df, column: str, label: Any) -> None:
+    def __init__(self, index, df, column: str, label: Any) -> None:
+        self.index = index
         self.df = df
         self.column = column
         self.label = label
+    
+    @property
+    def labels(self):
+        return {self.column: self.label}
 
     def apply_selection(self, df):
         raise NotImplementedError
 
-    def execute(self):
-        df = self.apply_selection(self.df)
-        if df.index.names:
-            df = df.reset_index()
-        return df.to_dict(orient="records")
+    def execute(self, limit: int = None, skip: int = None):
+        logger.debug('Applying pandas dataframe selection')
 
+        if not len(self.df):
+            return []
+        df = self.apply_selection(self.df)
+       
+        if df.index.names or df.index.name:
+            df = df.reset_index()
+        if limit is not None:
+            start = skip * self.index.DOCS_PER_LABEL if skip is not None else 0
+            limit = start + limit * self.index.DOCS_PER_LABEL
+            df = df.iloc[start:limit]
+        docs = df.to_dict(orient="records")
+        docs = self.index.reduce(docs, self.labels)
+        logger.debug(f'Done. Found {len(docs)} documents.')
+        return docs
+
+    def min(self, fields: Union[str,List[str]]):
+        if isinstance(fields, str):
+            fields = [fields]
+        df = self.apply_selection(self.df)
+        results = {}
+        for field in fields:
+            if field in df.index.names:
+                df = df.reset_index()
+            results[field] = df[field].min()
+        if len(fields) == 1:
+            return results[fields[0]]
+        return results
+
+    def max(self, fields: Union[str,List[str]]):
+        if isinstance(fields, str):
+            fields = [fields]
+        df = self.apply_selection(self.df)
+        results = {}
+        for field in fields:
+            if field in df.index.names:
+                df = df.reset_index()
+            results[field] = df[field].max()
+        if len(fields) == 1:
+            return results[fields[0]]
+        return results
+    
+    def unique(self, fields: Union[str,List[str]]):
+        if isinstance(fields, str):
+            fields = [fields]
+        df = self.apply_selection(self.df)
+        results = {}
+        for field in fields:
+            if field in df.index.names:
+                df = df.reset_index()
+            results[field] = list(df[field].unique())
+        if len(fields) == 1:
+            return results[fields[0]]
+        return results
+
+    def count(self):
+        df = self.apply_selection(self.df)
+        return len(df)
 
 class PandasSimpleQuery(PandasBaseQuery):
     def apply_selection(self, df):
@@ -101,22 +161,44 @@ class PandasInterpolationQuery(PandasBaseQuery):
             # same as before
             after = after.sort_values(self.column, ascending=True).head(limit)
             rows.append(after)
-
+        if not rows:
+            return df.head(0)
         return pd.concat(rows)
 
 
 class PandasMultiQuery(PandasBaseQuery):
-    def __init__(self, queries: List[PandasBaseQuery]) -> None:
+    def __init__(self, index, df, queries: List[PandasBaseQuery]) -> None:
+        self.index = index
+        self.df = df
         self.queries = queries
 
+    @property
+    def labels(self):
+        return {query.column: query.label for query in self.queries}
+
     def apply_selection(self, df):
+        if len(self.queries) == 1:
+            return self.queries[0].apply_selection(df)
+
         for query in self.queries:
             if isinstance(query, PandasInterpolationQuery):
                 selections = []
                 others = [q.column for q in self.queries if q is not query]
+                if not others:
+                    df = query.apply_selection(df)
+                    continue
+
                 for _, pdf in df.groupby(others):
-                    selections.append(query.apply_selection(pdf).reset_index())
-                df = pd.concat(selections)
+                    selection = query.apply_selection(pdf).reset_index()
+                    selections.append(selection)
+
+                selections = [s for s in selections if len(s)]
+                if not selections:
+                    df = df.head(0)
+                elif len(selections) == 1:
+                    df = selections[0]
+                else:
+                    df = pd.concat(selections)
             else:
                 df = query.apply_selection(df)
         return df
@@ -138,7 +220,7 @@ class PandasInterface(DatasourceInterface):
             return cls(df)
         
         raise NotImplementedError
-        
+
     @singledispatchmethod
     def compile_query(self, index, label):
         raise NotImplementedError(
@@ -147,26 +229,35 @@ class PandasInterface(DatasourceInterface):
 
     @compile_query.register(Index)
     def simple_query(self, index, label):
-        return PandasSimpleQuery(self.source, index.name, label)
+        return PandasSimpleQuery(index, self.source, index.name, label)
 
     @compile_query.register(IntervalIndex)
     def interval_query(self, index, label):
-        return PandasIntervalQuery(self.source, index.name, label)
+        return PandasIntervalQuery(index, self.source, index.name, label)
 
     @compile_query.register(InterpolatingIndex)
     def interpolating_query(self, index, label):
-        return PandasInterpolationQuery(self.source, self.source, index.name, label)
+        return PandasInterpolationQuery(index, self.source, index.name, label)
 
     @compile_query.register(list)
     @compile_query.register(tuple)
     @compile_query.register(MultiIndex)
-    def multi_query(self, indexes, labels):
-        if isinstance(indexes, MultiIndex):
-            indexes = indexes.indexes
-            labels = labels.values()
+    def multi_query(self, index, labels):
+        if not isinstance(index, MultiIndex):
+            index = MultiIndex(*index)
 
         queries = [
-            self.compile_query(idx, label) for idx, label in zip(indexes, labels)
+            self.compile_query(idx, labels[idx.name]) for idx in index.indexes
         ]
 
-        return PandasMultiQuery(self.source, queries)
+        return PandasMultiQuery(index, self.source, queries)
+
+    def insert(self, doc):
+        schema = doc.__class__
+        index = tuple(schema.index_for(name).to_pandas(label)
+                        for name, label in doc.index_labels.items())
+        if len(index) == 1:
+            index = index[0]
+        self.source.loc[index,:] = doc.column_values
+
+    update = insert
